/**** SimCore since 2002-05-15 by Kenji KISE ****/ 
/************************************************/
#include <unistd.h> 
#include "scalar.h"

/** system_config for RealScalar               **/
/************************************************/
system_config_s::system_config_s(char **opt){
  btb_entry      = 512;
  icache_entry   = 512;
  icache_penalty = 8;
  dcache_entry   = 512;
  dcache_penalty = 8;
  branch_miss_penalty = 2;
  P_INFETCH = 0;
  P_IDECODE = 1;
  P_REGREAD = 2;
  P_EXECUTE = 3;
  P_MEM_ACC = 4;
  P_WRITEBK = 5;
  P_DEPTH   = 6; /* pipeline depth */

  if(opt==NULL) return;
  /***** parse options *****/
  for(int i=0; opt[i]!=NULL; i++){
    char buf[1024];
    strcpy(buf, opt[i]);
    if(buf[0]!='-') continue;
    switch(buf[1]){
    case 'C': 
      icache_penalty = atoi(&buf[2]);
      dcache_penalty = atoi(&buf[2]);
      printf("** cache latency %d\n", 
	     icache_penalty);
      break;
    case 'P': 
      P_DEPTH = atoi(&buf[2]);
      P_REGREAD = P_DEPTH-4;
      P_EXECUTE = P_DEPTH-3;
      P_MEM_ACC = P_DEPTH-2;
      P_WRITEBK = P_DEPTH-1;
      printf("** pipeline depth %d\n", P_DEPTH);
      if(P_DEPTH<6){
	printf("** pipeline is too short.\n");
	exit(1);
      }
      if(P_DEPTH>=BUF_SIZE){
	printf("** pipeline is too long.\n");
	exit(1);
      }
      break;
    }
  }
}

/** return the instruction latency             **/
/************************************************/
int scalar_chip::latency(char Op, int CI){
  if(CI==_AND___) return 10;
  if(Op==OP_INTM) return 4;
  if(Op==OP_FLTI) return 5;
  return 1; /* default instruction latency */
}

/** print usage for SimCore_RealScalar         **/
/************************************************/
void usage_rs(){
  char UsageMessage[] =  "\
Usage: SimCore_RealScalar [-option] command\n\
 -d: debug mode\n\
 -z: verify mode\n\
 -a: print simple evaluation result\n\
 -c: print systemcall trace\n\
 -t[num][m]: print execution trace\n\
 -e[num][m]: end simulation at num code run. \
m means million\n\
 -f[num]: set fd_adjust to verify simulation\n\
 -v[num]: verbose mode, prints '.'\
by VERBOSE_INTERVAL\n\
 -P[num]: set pipeline depth\n\
 -C[num]: set cache latency\n\
 -i[file_name]: file_name as STDIN input file\n\
";
  printf("\n%s", UsageMessage);
}

/************************************************/
void scalar_chip::print_result(){
  class evaluation_result   *e  = ev->e;
  printf("===\n");
  printf("=== Pipeline depth       : %11d\n", 
	 ev->sc2->P_DEPTH);
  printf("=== BTB entry            : %11d\n", 
         ev->sc2->btb_entry);
  printf("=== Branch_Miss_Penalty  : %11d\n", 
         ev->sc2->branch_miss_penalty);
  printf("=== Conditional   Branch : %11lld\n",
         ev->e2->c_branch);
  
  printf("=== Cond Branch Hit      : %11lld\n",
         ev->e2->c_branch_hit);
  printf("=== Cond Branch Miss     : %11lld\n",
         ev->e2->c_branch_mis);
  printf("=== Cond Branch Hit Rate : %11.3f\n",
         (double)ev->e2->c_branch_hit / 
         (double)
	 (ev->e2->c_branch_hit + 
	  ev->e2->c_branch_mis));
  printf("===\n");
  
  printf("=== UnConditional Branch : %11lld\n",
         ev->e2->unc_branch);
  printf("===\n");
  
  uint_64t icache_hit = 
    ev->e2->icache_access - ev->e2->icache_mis;
  printf("=== Icache entry         : %11d\n",
         ev->sc2->icache_entry);
  printf("=== Icache Miss Penalty  : %11d\n",
         ev->sc2->icache_penalty);
  printf("=== Icache Access        : %11lld\n",
         ev->e2->icache_access);
  printf("=== Icache Hit           : %11lld\n",
         icache_hit);
  printf("=== Icache Miss          : %11lld\n",
         ev->e2->icache_mis);
  printf("=== Icache Hit Rate      : %11.3f\n",
         (double)icache_hit / 
         (double)(ev->e2->icache_access));
  printf("===\n");

  printf("=== Dcache entry         : %11d\n",
         ev->sc2->dcache_entry);
  printf("=== Dcache Miss Penalty  : %11d\n",
         ev->sc2->dcache_penalty);

  uint_64t dcache_hit = 
    ev->e2->dcache_access - ev->e2->dcache_mis;
  printf("=== Dcache Access        : %11lld\n",
         ev->e2->dcache_access);
  printf("=== Dcache Hit           : %11lld\n",
         dcache_hit);
  printf("=== Dcache Miss          : %11lld\n",
         ev->e2->dcache_mis);
  printf("=== Dcache Hit Rate      : %11.3f\n",
         (double)(dcache_hit)/ 
         (double)(ev->e2->dcache_access));
  printf("===\n");
  
  printf("=== Clock Cycles         : %11lld\n",
         ev->e2->clock);
  printf("=== CPI(cycles per inst.): %11.3f\n", 
         (double)ev->e2->clock/
	 (double)e->retired_inst);
  fflush(stdout);
}

#define FLYING_CODE_MASK 0xfff  /** 4095 insn. **/
/************************************************/
InstBuf::InstBuf(Env *ev){ /* Constructor */
  buf = new instruction_s*[FLYING_CODE_MASK+1];
  for(int i=0; i<=FLYING_CODE_MASK; i++){
    buf[i] = new instruction_s(ev);
  }
}

/************************************************/
instruction_s *InstBuf::alloc(){
  static int head = 0;
  head++;
  head &= FLYING_CODE_MASK;
  return buf[head];
}

#define WD_INT 0xff /** watch dog interval **/
/************************************************/
void scalar_chip::house_keeping(){
  system_config       *sc  = ev->sc;
  evaluation_result   *e   = ev->e;
  evaluation_result_s *e2  = ev->e2;
  static int wd_cnt = 0; /* watch dog count */

  /*** watch dog timer check by 0xff cycles ***/
  if((e2->clock & WD_INT)==WD_INT){
    if(wd_cnt==0){
      printf("\nWatch Dog Timer: retired %lld\n",
             e->retired_inst);
      exit(1);
    }
    wd_cnt=0;
  }

  if(sc->trace_flag!=0 &&
     sc->trace_flag<=(e->retired_inst+1))
    print_pipeline();

  instruction_s *pt = stg[ev->sc2->P_WRITEBK];
  /** write back stage has a code to retire **/
  if(pt!=Inull & stall==0){
    e->retired_inst++;
    wd_cnt++;

    if(sc->verify_flag) verify();
    if(sc->debug_flag) ev->deb->debugmode(pt);

    if(sc->verbose_interval && 
       (e->retired_inst % 
	sc->verbose_interval)==0){
      write(1, ".", 1);
    }

    if(sc->end_tc && e->retired_inst>=sc->end_tc)
      ev->sys->running = 0;
    
    /**** branch instruction_s *****/
    if(pt->BR || pt->Op==OP_JSR){
      /*** un-conditional branch ***/
      if(pt->Op==OP_JSR || 
	 pt->CI==_BR__ || pt->CI==_BSR_){
	e2->unc_branch++;
      }
      /*** conditional branch    ***/
      else{
	e2->c_branch++;
	if(pt->Ppc==pt->Npc) e2->c_branch_hit++;
	else                 e2->c_branch_mis++;
      }
    }
  } 
}

/* print pipeline state if -t option is used.  */
/************************************************/
void scalar_chip::print_pipeline(){
  if(ev->e2->clock==1){
    printf(" cycle| ");
    for(int i=0; i<ev->sc2->P_DEPTH; i++){
      if(i==sc2->P_INFETCH)      printf(" IF  "); 
      else if(i==sc2->P_IDECODE) printf(" ID  ");
      else if(i==sc2->P_REGREAD) printf(" RR  ");
      else if(i==sc2->P_EXECUTE) printf(" EX  ");
      else if(i==sc2->P_MEM_ACC) printf(" MM  ");
      else if(i==sc2->P_WRITEBK) printf(" WB  ");
      else printf("     ");
    }
    printf("|la:D:IC:DC|id|\n");
  }
  printf("%06lld| ", ev->e2->clock);

  if(stall>0){
    printf("pipeline stall | %d\n", stall);
    return;
  }

  for(int i=0; i<sc2->P_DEPTH; i++){
    if(stg[i]==Inull) printf("---- ");
    else{
      printf("%04llx ", 
             stg[i]->Cpc&0xffff);
    }
  }
  if(w->code_stall) printf("|%2d:",insn_latency);
  else              printf("|--:");

  if(w->data_stall) printf("d:");
  else              printf("-:");

  if(w->icache_stall) 
    printf("%2d:", w->icache_stall);
  else printf("--:");

  if(w->dcache_stall) 
    printf("%2d|", w->dcache_stall);
  else printf("--|");

  if(stg[sc2->P_WRITEBK]->LD)      printf("ld");
  else if(stg[sc2->P_WRITEBK]->ST) printf("st");
  else if(stg[sc2->P_WRITEBK]->CM) printf("cm");
  else if(stg[sc2->P_WRITEBK]->BR) printf("br");
  else printf("  ");
  printf("|");
  
  if(stg[sc2->P_WRITEBK]->ir){
    printf("%04lld ", ev->e->retired_inst + 1);
    int id = stg[sc2->P_WRITEBK]->CI;
    printf("%02x.%04x", id>>16, id & 0xffff);
  }
  printf("\n");
  
  if(w->flush_bpr){
    printf("Branch_Miss_Prediction %05lld", 
           ev->e2->bpred_miss + 1);
    printf(" (%04llx -> %04llx)\n\n", 
           stg[sc2->P_EXECUTE]->Ppc & 0xffff,
           stg[sc2->P_EXECUTE]->Npc & 0xffff);
  }
}

/************************************************/
void scalar_chip::verify(){
  simple_chip *c = s_chip;
  ADDR_TYPE c_cpc = c->ev->as->rr(0);
  c->step();
  instruction   *cp = c->p;
  instruction_s *sp = stg[ev->sc2->P_WRITEBK];

  int error = 0;
  for(int i=1; i<=64; i++)
    if(c->ev->as->rr(i)!=ev->as->rr(i)) error=1;

  if(c_cpc!=stg[sc2->P_WRITEBK]->Cpc) error=1;
  if((sp->LD || sp->ST) & cp->Adr != sp->Adr) 
    error=1;
  if(cp->Rav != sp->Rav) error=1;
  if(cp->Rbv != sp->Rbv) error=1;
  if((!sp->ST)& cp->Rcv != sp->Rcv) error=1;

  if(error){
    printf("** Verify Error:% lld code retire\n", 
           ev->e->retired_inst);
    printf("** simple_chip / real_scalar\n\n");
    printf("Cpc:%16llx %16llx\n", 
           c_cpc,   sp->Cpc);
    printf("Npc:%16llx %16llx\n", 
	   cp->Npc, sp->Npc);
    printf("Rav:%16llx %16llx\n", 
           cp->Rav, sp->Rav);
    printf("Rbv:%16llx %16llx\n", 
           cp->Rbv, sp->Rbv);
    printf("Rcv:%16llx %16llx\n", 
           cp->Rcv, sp->Rcv);
    printf("Adr:%16llx %16llx\n", 
           cp->Adr, sp->Adr);

    for(int i=1; i<=32; i++){
      print_register_name(i-1);
      printf("%02d:%016llx %016llx", i-1, 
             c->ev->as->rr(i), ev->as->rr(i));
      if(c->ev->as->rr(i)!=ev->as->rr(i)) 
        printf("*");
      else 
        printf(" ");
      printf(" f:%016llx %016llx", 
             c->ev->as->rr(i+32), 
             ev->as->rr(i+32));
      if(c->ev->as->rr(i+32)!=ev->as->rr(i+32))
        printf("*");
      else 
        printf(" ");
      printf("\n");
    }
    exit(1);
  }
}

/************************************************/
scalar_chip::scalar_chip(char *prog, char **opt){
  ev       = new Env_s();
  ev->sc   = new system_config(prog, opt);
  ev->e    = new evaluation_result();
  ev->as   = new architecture_state(ev);
  ev->mem  = new memory_system(ev);
  ev->deb  = new debug(ev);
  ev->sys  = new system_manager(ev);

  ev->sc2  = new system_config_s(opt);
  sc2      = ev->sc2;
  ev->e2   = new evaluation_result_s();

  pc_gen   = new PCgen(ev);
  b_pred   = new Bpred(ev);
  icache   = new Icache(ev);
  dcache   = new Dcache(ev);
  inst_buf = new InstBuf(ev);
  w        = new chip_wire();

  Inull    = new instruction_s(ev);
  Inull->ir = 0;
  Inull->Slot();                 

  stg   = new instruction_s*[sc2->P_DEPTH];
  w->stg= new instruction_s*[sc2->P_DEPTH];
  for(int i=0; i<sc2->P_DEPTH; i++){
    w->stg[i] = stg[i] = Inull;
  }
  s_chip = NULL;

  stall        = 0;
  insn_latency = 1;
}

/************************************************/
scalar_chip::~scalar_chip(){   /** destructor **/
  print_evaluation_result(ev);
  if(!ev->sc->simple_result) print_result();
  if(ev->sc->result_flag){
    data_t a, d;
    while(ev->mem->mm->load_all(&a, &d)){
      if(a>=0x40000000) 
        printf("1%llx %16llx\n", a, d);
    }
  }
}

/************************************************/
void scalar_chip::Init(){
  w->pcgen_pc = pc_gen->pc;
  
  w->stg[sc2->P_INFETCH] = Inull;
  for(int i=1; i<sc2->P_DEPTH; i++) 
    w->stg[i]=stg[i-1];
  
  instruction_s *p1 = w->stg[sc2->P_WRITEBK];
  w->dc_wir  = (p1->ST) ? p1->ir  : 0;
  w->dc_wadr = (p1->ST) ? p1->Adr : 0;
  w->dc_wdat = (p1->ST) ? p1->Rav : 0;

  instruction_s *p2 = w->stg[sc2->P_MEM_ACC];
  w->dc_rir  = (p2->LD) ? p2->ir  : 0;
  w->dc_radr = (p2->LD) ? p2->Adr : 0;
  w->dc_rdat = 0;
}

/************************************************/
void scalar_chip::IDecode(){
  instruction_s *pt = w->stg[sc2->P_IDECODE];

  pt->Slot();
  pt->Issue();

  /*** a part of decode for dependency check ***/
  pt->Ra =(pt->Ar << 5) |  pt->RA;
  pt->Rb =(pt->Br << 5) |  pt->RB;
  pt->Rc =(pt->WB==31) ? 0xff:
    (WR(pt->Op)<< 5) | pt->WB;
  pt->Rm = (!pt->CM) ? 0xf0 : 
    (WR(pt->Op)<< 5) | pt->WB;
}

/************************************************/
void scalar_chip::RegRead(){
  w->stg[sc2->P_REGREAD]->RegisterRead();
}

/************************************************/
void execute_one(instruction_s *p, chip_wire *w){
  p->Execute();
  data_t target_pc = p->ev->as->pc;
  int mis = (p->Ppc!=target_pc);
  w->flush_bpr = (mis) ? 1 : 0;
  w->pcgen_bpr = (mis) ? target_pc : 0;
}

/************************************************/
void scalar_chip::Execute(){
  static int busy=0; /* ALU is busy ? */
  w->code_stall = 0;
  w->flush_bpr  = 0;
  w->pcgen_bpr  = 0;
    
  instruction_s *pt = w->stg[sc2->P_EXECUTE];
  if(pt==Inull) return;

  if(!busy){ /** fire new instruction  **/
    insn_latency = latency(pt->Op, pt->CI);
    if(insn_latency==1){
      execute_one(pt, w);
      busy=0;
    }
    else{
      busy=1;
      w->code_stall=1;
    }
  }
  else{ /** executing multi-cycle inst. **/
    if(insn_latency>1) insn_latency--;
    if(insn_latency==1){
      execute_one(pt, w);
      busy=0;
    }
    else{
      w->code_stall=1;
    }
  }
}

/************************************************/
void scalar_chip::WriteBack(){
  instruction_s *p = w->stg[sc2->P_WRITEBK];

  int br = (p->Op==OP_JSR || p->BR);
  w->bpred_wtkn = (br) & 
    (p->Cpc+4 != p->Npc);
  w->bpred_wpc  = (br) ? p->Cpc : 0;
  w->bpred_wnpc = (br) ? p->Npc : 0;
  
  /***** Data Cache Flush by PAL*****/
  if(p!=Inull && p->Op==OP_PAL) dcache->flush(); 

  if(p!=Inull){
    ev->as->pc = p->Npc;
    p->WriteBack();  /** Write Back **/
  }
  
  int pal = (p!=Inull && p->Op==OP_PAL);
  /* sigreturn updates as->ps */
  w->pcgen_pal = (pal) ? ev->as->pc : 0;
  w->flush_pal = (pal) ? 1 : 0;
}

/************************************************/
void scalar_chip::Step(){
  /***** data deps stall calculation *****/
  if(w->stg[sc2->P_REGREAD]==Inull || 
     w->stg[sc2->P_EXECUTE]==Inull){
    w->data_stall = 0;
  }
  else{
    instruction_s *pr = w->stg[sc2->P_REGREAD];
    instruction_s *pw = w->stg[sc2->P_EXECUTE];
    
    /** dependency: RegRead after Load **/
    int w_ld  = pw->LD && pw->WB!=31 &&
      (pr->Ra == pw->Rc || pr->Rb == pw->Rc);
    
    /** depencency: CMOV    after Load **/
    int w_cm = pr->CM && pw->LD && pw->WB!=31 &&
      pr->Rm == pw->Rc;

    w->data_stall = w_ld | w_cm;
  }
}

/************************************************/
void scalar_chip::UpdatePipeReg(){
  char null[BUF_SIZE];
  char updt[BUF_SIZE];
  for(int i=0; i<sc2->P_DEPTH; i++){
    null[i] = 0;
    updt[i] = 1;
  }
  
  /***** (1) Flush by PAL Code        *****/
  if(w->flush_pal){
    for(int i=0; i<sc2->P_WRITEBK; i++)
      null[i]=1;
  }
  /***** (2) Stall by D-cache Miss    *****/
  else if(w->dcache_stall){
    null[sc2->P_MEM_ACC]=1;
    for(int i=0; i<sc2->P_MEM_ACC; i++) 
      updt[i]=0;
  }
  /***** (3) Flush by Branch Mis-pred *****/
  else if(w->flush_bpr){
    for(int i=0; i<sc2->P_EXECUTE; i++)
      null[i]=1;
    stall = sc2->branch_miss_penalty;
    ev->e2->bpred_miss++; /** evaluation **/
  }
  /***** (4) Stall by Code Latency    *****/
  else if(w->code_stall){
    null[sc2->P_INFETCH]=1;
    null[sc2->P_EXECUTE]=1;
    for(int i=0; i<sc2->P_EXECUTE; i++) 
      updt[i]=0;
  }
  /***** (5) Stall by Data Dependence *****/
  else if(w->data_stall){
    null[sc2->P_INFETCH]=1;
    null[sc2->P_REGREAD]=1;
    for(int i=0; i<sc2->P_REGREAD; i++)
      updt[i]=0;
  }
  
  for(int i=0; i<sc2->P_DEPTH; i++){
    if(null[i]) w->stg[i] = Inull;
    if(updt[i]) stg[i] = w->stg[i];
  }
}

/************************************************/
void scalar_chip::Update(){
  /***** Set fetched code info. *****/
  if(!w->icache_stall){
    instruction_s *p = inst_buf->alloc();
    p->ir  = w->ic_out;
    p->Cpc = w->pcgen_pc;
    p->Npc = w->pcgen_npc;
    p->Ppc = (w->bpred_tkn) ? w->bpred_targ :
      w->pcgen_npc;
    w->stg[sc2->P_INFETCH] = p;
  }

  /***** Set the Rcv from the data cache *****/
  if(w->stg[sc2->P_MEM_ACC]->LD) 
    w->stg[sc2->P_MEM_ACC]->Rcv=w->dc_rdat;

  /**** Data forwarding to Rav & Rbv & Adr ****/
  if(w->stg[sc2->P_REGREAD]!=Inull){
    instruction_s *pr = w->stg[sc2->P_REGREAD];
    for(int i=sc2->P_WRITEBK; 
	i>=sc2->P_EXECUTE; i--){
      instruction_s *pw = w->stg[i];
      if(pw!=Inull){
	if(pr->Ra==pw->Rc) pr->Rav=pw->Rcv;
	if(pr->Rb==pw->Rc) pr->Rbv=pw->Rcv;
	if(pr->Rm==pw->Rc) pr->Adr=pw->Rcv;
      }
    }
  }

  int hazard = w->flush_pal | w->dcache_stall | 
    w->flush_bpr | w->code_stall | w->data_stall;
  if(!hazard){
    for(int i=0; i<sc2->P_DEPTH; i++){
      stg[i] = w->stg[i];
    }
  }
  else {
    UpdatePipeReg();
  }
}

/************************************************/
void scalar_chip::main_loop(){
  while(ev->sys->running){
    if(stall) stall--;
    else{
      /******** (1) Init    ********/
      Init();
      
      /******** (2) Step    ********/
      Step();
      pc_gen->Step(w); /* IF */
      b_pred->Step(w); /* IF */
      icache->Step(w); /* IF */
      dcache->Step(w); /* MM */
      IDecode();       /* ID */
      RegRead();       /* RR */
      Execute();       /* EX */
      WriteBack();     /* WB */
      
      /******** (3) Update  ********/
      Update();
      pc_gen->Update(w);
      b_pred->Update(w);
      icache->Update(w);
      dcache->Update(w);
    }
    /******** etc.        ********/
    ev->e2->clock++; 
    house_keeping();
  }
}

/************************************************/
int main(int argc, char **argv){
  printf("%s/%s      %s\n", PROJ,RS_NAME,RS_VER);
  printf("on %s %s\n", NAME, VER); 
  if(argc==1) usage_rs();
  else{
    char *p    = argv[argc-1]; /* prog name */
    char **opt = argv;         /* options   */
    scalar_chip *c = new scalar_chip(p, opt);
    
    if(c->ev->sc->verify_flag){
      char *op[OPTION_MAX];
      option_from_file(op, "verify.txt");
      c->s_chip = new simple_chip(p, op);
    }
    
    c->main_loop();
    delete c;
  }
  return 0;
}
/************************************************/
